"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateMiniLogoCmds = void 0;
const ast_1 = require("../language-server/generated/ast");
/**
 * Generates simple drawing commands from a MiniLogo Model
 * @param model Model to generate commmands from
 * @returns Generated commands that captures the program's drawing intent
 */
function generateMiniLogoCmds(model) {
    return generateStatements(model.stmts);
}
exports.generateMiniLogoCmds = generateMiniLogoCmds;
function generateStatements(stmts) {
    // setup drawing state
    let state = {
        px: 0,
        py: 0,
        drawing: false
    };
    // minilogo evaluation env
    let env = new Map();
    // generate mini logo cmds off of statements
    return stmts.flatMap(s => evalStmt(s, env, state)).filter(e => e !== undefined);
}
/**
 * Takes an env, a drawing state, and the active file node we're appending to
 * Effectful & recursive statement evaluation
 */
function evalStmt(stmt, env, state) {
    var _a, _b, _c;
    if ((0, ast_1.isPen)(stmt)) {
        if (stmt.mode === 'up') {
            state.drawing = false;
            return [{
                    cmd: 'penUp'
                }];
        }
        else {
            state.drawing = true;
            return [{
                    cmd: 'penDown'
                }];
        }
    }
    else if ((0, ast_1.isMove)(stmt)) {
        // update pen position
        let cmds = [{
                cmd: 'move',
                x: evalExprWithEnv(stmt.ex, env),
                y: evalExprWithEnv(stmt.ey, env)
            }];
        if (state.drawing) {
            cmds.push({ cmd: 'penUp' });
            cmds.push({ cmd: 'penDown' });
        }
        else {
            cmds.push({ cmd: 'penDown' });
            cmds.push({ cmd: 'penUp' });
        }
        return cmds;
    }
    else if ((0, ast_1.isMacro)(stmt)) {
        // get the cross ref & validate it
        const macro = stmt.def.ref;
        if (macro === undefined) {
            throw new Error((_b = (_a = stmt.def.error) === null || _a === void 0 ? void 0 : _a.message) !== null && _b !== void 0 ? _b : `Attempted to reference an undefined macro: ${stmt.def.$refText}`);
        }
        // copied env
        let macroEnv = new Map(env);
        // produce pairs of string & exprs, using a tmp env
        // this is important to avoid mixing of params that are only present in the tmp env w/ our actual env
        let tmpEnv = new Map();
        macro.params.map((elm, idx) => tmpEnv.set(elm.name, evalExprWithEnv(stmt.args[idx], macroEnv)));
        tmpEnv.forEach((v, k) => macroEnv.set(k, v));
        return macro.body.flatMap(s => evalStmt(s, macroEnv, state));
    }
    else if ((0, ast_1.isFor)(stmt)) {
        // for loop bounds
        let vi = evalExprWithEnv(stmt.e1, env);
        let ve = evalExprWithEnv(stmt.e2, env);
        let results = [];
        // perform loop
        const loopEnv = new Map(env);
        while (vi < ve) {
            loopEnv.set(stmt.var.name, vi++);
            stmt.body.forEach(s => {
                results = results.concat(evalStmt(s, new Map(loopEnv), state));
            });
        }
        return results;
    }
    else if ((0, ast_1.isColor)(stmt)) {
        // apply color to stroke
        if (stmt.color) {
            return [{ cmd: 'color', color: stmt.color }];
        }
        else {
            const r = evalExprWithEnv(stmt.r, env);
            const g = evalExprWithEnv(stmt.g, env);
            const b = evalExprWithEnv(stmt.b, env);
            return [{ cmd: 'color', r, g, b }];
        }
    }
    else {
        throw new Error((_c = 'Unrecognized Statement encountered: ' + (stmt === null || stmt === void 0 ? void 0 : stmt.$type)) !== null && _c !== void 0 ? _c : 'Unknown Type');
    }
}
// Evaluates exprs to final vals for emission
function evalExprWithEnv(e, env) {
    var _a, _b, _c, _d;
    if ((0, ast_1.isLit)(e)) {
        return e.val;
    }
    else if ((0, ast_1.isRef)(e)) {
        const v = env.get((_b = (_a = e.val.ref) === null || _a === void 0 ? void 0 : _a.name) !== null && _b !== void 0 ? _b : '');
        if (v !== undefined) {
            return v;
        }
        throw new Error((_d = (_c = e.val.error) === null || _c === void 0 ? void 0 : _c.message) !== null && _d !== void 0 ? _d : `Attempted to lookup an unbound reference '${e.val.$refText}' in the env.`);
    }
    else if ((0, ast_1.isBinExpr)(e)) {
        let opval = e.op;
        let v1 = evalExprWithEnv(e.e1, env);
        let v2 = evalExprWithEnv(e.e2, env);
        switch (opval) {
            case '+': return v1 + v2;
            case '-': return v1 - v2;
            case '*': return v1 * v2;
            case '/': return v1 / v2;
            default: throw new Error(`Unrecognized bin op passed: ${opval}`);
        }
    }
    else if ((0, ast_1.isNegExpr)(e)) {
        return -1 * evalExprWithEnv(e.ne, env);
    }
    else if ((0, ast_1.isGroup)(e)) {
        return evalExprWithEnv(e.ge, env);
    }
    throw new Error('Unhandled Expression: ' + e);
}
//# sourceMappingURL=generator.js.map